import tensorflow as tf
import tensorflow_probability as tfp
import keras.backend as K
tfd = tfp.distributions


class Losses:
    def __init__(self, loss_fn, observation_noise, beta=0,
                 z_mean=None, z_sigma=None, z2_mean=None, z2_log_sigma=None,
                 z_mean_TD=None, z_log_sigma_TD=None):
        self.loss = self._get_loss(loss_fn)
        self.beta = beta
        self.observation_noise = observation_noise
        self.z2_mean = z2_mean
        self.z2_log_sigma = z2_log_sigma

        def f1():
            assert z2_mean != None, "z_mean should be defined!"
            assert z2_log_sigma != None, "z_log_sigma should be defined"
            return 0

        def f2():
            return 0
        tf.cond(self.beta > 0, f1, f2)
        self.z_mean = z_mean
        self.z_sigma = z_sigma
        self.z_mean_TD = z_mean_TD
        self.z_log_sigma_TD = z_log_sigma_TD
        if self.z_mean != None and self.z_sigma != None:
            assert z_mean_TD != None, "z_mean_TD should be defined!"
            assert z_log_sigma_TD != None, "z_log_sigma_TD should be defined!"

    def _get_loss(self, loss_fn):
        losses = {"binary": self._binary,
                  "normal": self._normal,
                  "normalDiag": self._normalDiag,
                  "bernoulli": self._bernoulli}
        return losses[loss_fn]

    """
      Making it custom metric to be able to feed it to Keras API - actually no need for y_true, y_pred
    """

    def KL_divergence(self, y_true, y_pred):
        def f1():
            p2 = tfd.Normal(self.z2_mean, tf.exp(self.z2_log_sigma) + 1e-12)
            q2 = tfd.Normal(0, 1)
            if self.z_mean != None and self.z_sigma != None:
                p = tfd.Normal(self.z_mean, self.z_sigma + 1e-12)
                q = tfd.Normal(self.z_mean_TD, tf.exp(
                    self.z_log_sigma_TD) + 1e-12)
                return self.beta * (tf.reduce_mean(tfd.kl_divergence(p, q)) + tf.reduce_mean(tfd.kl_divergence(p2, q2)))
            else:
                return self.beta * tf.reduce_mean(tfd.kl_divergence(p2, q2))

        def f2(): return 0.0
        return tf.cond(self.beta > 0., f1, f2)
    """
      For binarized input with KL term (!)
    """

    def _binary(self, x_true, x_reco):
        return -tf.nn.sigmoid_cross_entropy_with_logits(
            labels=x_true, logits=x_reco) + self.KL_divergence(None, None)

    def _bernoulli(self, x_true, x_reco):
        return -tf.reduce_mean(tfd.Bernoulli(logits=x_reco)._log_prob(x_true)
                               ) + self.KL_divergence(None, None)

    """
      For non binarized input with KL term(!)
    """

    def _normal(self, x_true, x_reco):
        return -tf.reduce_mean(
            tfd.Normal(x_reco, scale=self.observation_noise)._log_prob(x_true)
        ) + self.KL_divergence(None, None)

    def _normalDiag(self, x_true, x_reco):
        return -tf.reduce_mean(
            tfd.MultivariateNormalDiag(
                x_reco, scale_identity_multiplier=self.observation_noise)._log_prob(x_true)) + self.KL_divergence(None, None)



from .encoder import Encoder
from keras.layers import Input, Conv2D, Conv2DTranspose, MaxPooling2D, PReLU, Dense, Flatten, Reshape
from keras.models import Model


class ConvolutionalEncoder(Encoder):

    def _encoder(self):
        input_img = Input(shape=self.input_shape[1:])
        x = Conv2D(512, (2, 2), padding='same')(input_img)
        x = PReLU()(x)
        x = MaxPooling2D((2, 2), padding='same')(x)
        x = Conv2D(256, (2, 2), padding='same')(x)
        x = PReLU()(x)
        x = MaxPooling2D((2, 2), padding='same')(x)
        x = Conv2D(128, (2, 2), padding='same')(x)
        x = PReLU()(x)
        x = MaxPooling2D((2, 2), padding='same')(x)
        encoded = Flatten()(x)
        encoder = Model(input_img, encoded, name="convolutional_encoder")
        return encoder

    def _decoder(self):
        latent = Input(shape=(self.latent_dim,))
        x = Reshape((4, 4, self.latent_dim // 16))(latent)
        x = Conv2D(256, (2, 2), padding='same')(x)
        x = PReLU()(x)
        x = Conv2DTranspose(256, (2, 2), strides=(2, 2), padding='same')(x)
        x = PReLU()(x)
        x = Conv2D(512, (2, 2))(x)
        x = PReLU()(x)
        x = Conv2D(1024, (2, 2), padding='same')(x)
        x = PReLU()(x)
        x = Conv2DTranspose(1024, (2, 2), strides=(2, 2), padding='same')(x)
        x = PReLU()(x)
        reco = Conv2DTranspose(self.input_shape[-1], (2, 2), strides=(2, 2))(x)
        decoder = Model(latent, reco, name="convolutional_decoder")
        return decoder



from .small_dense_lvae import SmallDenseLadderVAE
from keras.layers import Input, Dense, ReLU
from keras.models import Model


class SmallDenseLinearLadderVAE(SmallDenseLadderVAE):
    def encoder1(self):
        input_tensor = Input(shape=self.input_shape[1:])
        x = Dense(256)(input_tensor)
        x = Dense(self.latent_dim1)(x)
        encoder = Model(input_tensor, x, name="linear_encoder")
        return encoder



from .variational_autoencoder import VariationalAutoEncoder
from keras.layers import Input, Conv2D, ReLU, MaxPooling2D, Flatten, Reshape, Conv2DTranspose
from keras.models import Model


class SmallConvolutionalVAE(VariationalAutoEncoder):
    def _encoder(self):
        input_img = Input(shape=self.input_shape[1:])
        x = Conv2D(256, (2, 2), padding='same')(input_img)
        x = ReLU()(x)
        x = MaxPooling2D((2, 2), padding='same')(x)
        x = Conv2D(128, (2, 2), padding='same')(x)
        x = ReLU()(x)
        x = MaxPooling2D((2, 2), padding='same')(x)
        x = Conv2D(128, (2, 2), padding='same')(x)
        x = ReLU()(x)
        x = MaxPooling2D((2, 2), padding='same')(x)
        encoded = Flatten()(x)
        encoder_model = Model(input_img, encoded,
                              name="small_convolutional_encoder")
        return encoder_model

    def _decoder(self):
        latent = Input(shape=(self.latent_dim,))
        x = Reshape((4, 4, self.latent_dim // 16))(latent)
        x = Conv2DTranspose(128, (2, 2), strides=(2, 2), padding='same')(x)
        x = ReLU()(x)
        x = Conv2D(128, (2, 2))(x)
        x = ReLU()(x)
        x = Conv2DTranspose(512, (2, 2), strides=(2, 2), padding='same')(x)
        x = ReLU()(x)
        reco = Conv2DTranspose(
            self.input_shape[-1], (2, 2), strides=(2, 2), padding='same')(x)
        decoder_model = Model(latent, reco, name="small_convolutional_decoder")
        return decoder_model



import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import numpy as np
from keras.utils import to_categorical
from scipy.stats import pearsonr


class VAEPlotter:
    def __init__(self, fitted_model, datagen, label_datagen, grid_size=14):
        self.model = fitted_model.model
        self.generator_model = fitted_model.generator
        self._latent_dim = fitted_model.latent_dim
        self.datagen = datagen
        self.label_datagen = label_datagen
        self.grid_size = grid_size
        self.train_images, _ = next(self.datagen.flow())
        self.test_images, _ = self.datagen.validation_data()
        self.batch_size = self.train_images.shape[0]
        self.input_shape = self.model.layers[0].output_shape
        self.image_shape = self.train_images.shape[1:]
        if self.image_shape[2] == 1:
            self.image_shape = self.image_shape[:-1]

    def grid_plot(self):
        reco_train = self.model.predict(self.train_images.reshape(
            self.batch_size, *self.input_shape[1:]),
                                        batch_size=self.batch_size)

        reco_test = self.model.predict(
            self.test_images[:self.batch_size].reshape(self.batch_size,
                                                       *self.input_shape[1:]),
            batch_size=self.batch_size)

        reco = [reco_train, reco_test]
        images = [self.train_images, self.test_images]

        fig = plt.figure(figsize=(14, 7))
        outer = gridspec.GridSpec(1, 2, wspace=0.1, hspace=0.1)

        for outer_ind in range(2):
            inner = gridspec.GridSpecFromSubplotSpec(
                4,
                4,
                subplot_spec=outer[outer_ind],
                wspace=0.095,
                hspace=0.095)
            for inner_ind in range(16):
                ax = plt.Subplot(fig, inner[inner_ind])
                if inner_ind % 2 == 0:
                    if inner_ind < 4:
                        ax.set_title('Reconstructed')
                    ax.imshow(
                        reco[outer_ind][inner_ind].reshape(*self.image_shape),
                        interpolation='none',
                        vmin=0,
                        vmax=1)
                else:
                    if inner_ind < 4:
                        ax.set_title(' <- Original')
                    ax.imshow(images[outer_ind][inner_ind -
                                                1].reshape(*self.image_shape),
                              interpolation='none',
                              vmin=0,
                              vmax=1)
                ax.set_xticks([])
                ax.set_yticks([])
                fig.add_subplot(ax)

        fig.suptitle("Train and test reconstructions\n\n")
        plt.savefig("reconstrunction_samples.png", dpi=200)
        plt.show()

    def generate_samples(self, vmax=1):
        latent_inputs = np.random.normal(size=(self.grid_size * self.grid_size,
                                               self._latent_dim))

        self._plot_samples(latent_inputs, vmax)

    def visualize_latent(self, axis=0, sweep_from=-1, sweep_to=1, vmax=1):
        sweep = np.linspace(sweep_from, sweep_to,
                            self.grid_size * self.grid_size)
        latent_inputs = np.random.normal(size=(1, self._latent_dim))
        latent_inputs = np.array(latent_inputs.tolist() *
                                 (self.grid_size * self.grid_size))
        latent_inputs = latent_inputs.reshape(self.grid_size * self.grid_size,
                                              self._latent_dim)
        latent_inputs[:, axis] = sweep

        self._plot_samples(latent_inputs, vmax)

    def _plot_samples(self, latent_inputs, vmax):
        recos = self.generator_model.predict(
            latent_inputs[:self.grid_size * self.grid_size].reshape(
                self.grid_size * self.grid_size, self._latent_dim))
        # Output 196 images
        fig, axes = plt.subplots(self.grid_size,
                                 self.grid_size,
                                 sharex=True,
                                 sharey=True,
                                 figsize=(11, 11))

        if np.prod(recos[0].shape) / (28 * 28) != 1:
            recos = recos.reshape(self.grid_size * self.grid_size, 28, 28,
                                  int(np.prod(recos[0].shape) / (28 * 28)))
        else:
            recos = recos.reshape(self.grid_size * self.grid_size, 28, 28)

        for ind, ax in enumerate(axes.flatten()):
            ax.imshow(recos[ind], interpolation='none', vmin=0, vmax=vmax)
            ax.set_xticks([])
            ax.set_yticks([])

        fig.suptitle("Generated samples on %d - dimensional grid" %
                     self._latent_dim)
        plt.savefig("generated_samples.png", dpi=200)
        plt.show()

    def plot_label_correlations(self):
        images, labels = next(self.label_datagen.flow())
        recos, z1, z2 = self.model.predict(images, batch_size=self.batch_size)

        if np.prod(images[0].shape) / (28 * 28) != 1:
            images = images.reshape(self.batch_size, 28, 28,
                                    int(np.prod(images[0].shape) / (28 * 28)))
        else:
            images = images.reshape(self.batch_size, 28, 28)

        labels = to_categorical(labels)

        for cat in range(labels[0].shape[0]):
            fig = plt.figure(figsize=(10, 8))
            correlations = []

            for i in range(self.latent_dim):
                correlations.append(pearsonr(z2[:, i], labels[:, cat]))

            correlations = np.array(correlations)

            plt.title("#%d one-hot category correlation with z2" % (cat + 1))
            plt.scatter(range(correlations.shape[0]),
                        correlations[:, 0],
                        c=[
                            'b' if x >= 0.4 else 'r'
                            for x in np.abs(correlations[:, 0])
                        ])

            plt.hlines(y=.4, xmin=0, xmax=self._latent_dim)
            plt.hlines(y=-.4, xmin=0, xmax=self._latent_dim)
            plt.xlim(0, self._latent_dim)
            plt.xticks(np.arange(1, self._latent_dim, 2))
            plt.xlabel('Latent parameters')
            plt.ylabel('Pearson-correlation')
            plt.savefig('cat-%d-to-z2-corr.png' % (cat + 1), dpi=150)
            plt.show()



from keras.callbacks import Callback
import keras.backend as K


class IncrementalBeta(Callback):
    def __init__(self, beta, n_epochs=20):
        self.beta_max = beta
        self.n_epochs = n_epochs
        self.beta = 0

    def on_epoch_begin(self, epoch, logs=None):
        K.set_value(self.model.beta, self.beta)

    def on_epoch_end(self, epoch, logs=None):
        updated_beta = self.beta + self.beta_max / (3 * self.n_epochs // 4)
        if updated_beta < self.beta_max:
            self.beta = updated_beta
            print("Beta updated : ", self.beta)
        else:
            self.beta = self.beta_max



import pickle
from sklearn.utils import shuffle
from sklearn import model_selection
import numpy as np
import os
from sklearn.preprocessing import StandardScaler
from scipy.linalg import svd


class DataLoader:
    def __init__(self, file_path, binarize, zca_whiten=False):
        self.binarize = binarize
        self.zca_whiten = zca_whiten
        with open(file_path, 'rb') as f:
            self.data = pickle.load(f)

    def get_data_and_labels(self):
        X_train = self.data['train_images']
        X_test = self.data['test_images']
        y_train = self.data['train_labels']
        y_test = self.data['test_labels']

        X_train = np.clip(X_train, 0, 1.0)
        X_test = np.clip(X_test, 0, 1.0)

        print('Training size : %d \t Test size : %d' %
              (X_train.shape[0], X_test.shape[0]))

        channel = int(np.prod(X_train.shape) / (X_train.shape[0] * 28 * 28))

        # not possible to not split them
        X_train = shuffle(X_train.reshape(X_train.shape[0], 28, 28, channel),
                          random_state=42)
        y_train = shuffle(y_train, random_state=42)
        X_test = shuffle(X_test.reshape(X_test.shape[0], 28, 28, 1),
                         random_state=137)
        y_test = shuffle(y_test, random_state=42)

        if self.binarize:
            X_train[X_train >= .5] = 1.
            X_train[X_train < .5] = 0.
            X_test[X_test >= .5] = 1.
            X_test[X_test < .5] = 0.
        if self.zca_whiten:
            # add +1 and divide by 2 to shift
            # them back to the interval [-1, 1]
            X_train = self._whiten(X_train)
            X_test = self._whiten(X_test)

        print("Shapes : ", X_train.shape, "\t", X_test.shape)
        print("Label shaped : ", y_train.shape, "\t", y_test.shape)

        mean, std = np.mean(X_train), np.std(X_train)
        print("Train set : ")
        print('Mean: %.3f, Standard Deviation: %.3f' % (mean, std))
        print('Min: %.3f, Max: %.3f' % (np.min(X_train), np.max(X_train)))

        return (X_train, y_train, X_test, y_test)

    def train_test_split(self):
        try:
            X_train = self.data['train_images']
            X_test = self.data['test_images']

            X_train = np.clip(X_train, 0, 1.0)
            X_test = np.clip(X_test, 0, 1.0)

            print('Training size : %d \t Test size : %d' %
                  (X_train.shape[0], X_test.shape[0]))

            channel = int(
                np.prod(X_train.shape) / (X_train.shape[0] * 28 * 28))

            # not possible to not split them
            X_train = shuffle(X_train.reshape(X_train.shape[0], 28, 28,
                                              channel),
                              random_state=42)
            X_test = shuffle(X_test.reshape(X_test.shape[0], 28, 28, 1),
                             random_state=137)

            print("Shapes : ", X_train.shape, "\t", X_test.shape)
        except IndexError:
            X = self.data
            X = np.clip(X, 0, 1.0)

            channel = int(np.prod(X.shape) / (X.shape[0] * 28 * 28))

            X_train, X_test = model_selection.train_test_split(
                X.reshape(X.shape[0], 28, 28, channel),
                test_size=0.05,
                random_state=137)
        finally:
            if self.binarize:
                X_train[X_train >= .5] = 1.
                X_train[X_train < .5] = 0.
                X_test[X_test >= .5] = 1.
                X_test[X_test < .5] = 0.
            if self.zca_whiten:
                # add +1 and divide by 2 to shift
                # them back to the interval [-1, 1]
                X_train = self._whiten(X_train)
                X_test = self._whiten(X_test)

            mean, std = np.mean(X_train), np.std(X_train)
            print("Train set : ")
            print('Mean: %.3f, Standard Deviation: %.3f' % (mean, std))
            print('Min: %.3f, Max: %.3f' % (np.min(X_train), np.max(X_train)))

            return (X_train, X_test)

    def _whiten(self, x):
        flat_x = np.reshape(x,
                            (x.shape[0], x.shape[1] * x.shape[2] * x.shape[3]))
        sigma = np.dot(flat_x.T, flat_x) / flat_x.shape[0]
        u, s, _ = svd(sigma)
        s_inv = 1. / np.sqrt(s[np.newaxis] + 1e-12)
        principal_components = (u * s_inv).dot(u.T)
        flatx = np.reshape(x, (-1, np.prod(x.shape[-3:])))
        whitex = np.dot(flatx, principal_components)
        x = np.reshape(whitex, x.shape)
        return x
